from django.conf import settings


class TimepieceDefaults(object):

    TIMEPIECE_EXTRA_NAV = {}

    TIMEPIECE_DEFAULT_LOCATION_SLUG = None

    TIMEPIECE_PAID_LEAVE_PROJECTS = {}

    TIMEPIECE_ACCOUNTING_EMAILS = []

    TIMEPIECE_EMAILS_USE_HTTPS = True
